from django.core.management.base import BaseCommand
from bloomerp.utils.models import (
    get_create_view_url,
    get_list_view_url,
    get_detail_view_url,
    get_update_view_url,
    get_bulk_upload_view_url,
    get_model_dashboard_view_url,
    get_document_template_generate_view_url,
    get_document_template_list_view_url
)

class Command(BaseCommand):
    help = 'Generate links for all views in the application'

    def handle(self, *args, **options):
        pass


# Generated by Django 5.1.1 on 2024-12-22 21:16

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("test_app", "0025_employee_accommodations"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="employee",
            name="accommodations",
        ),
    ]



# Generated by Django 5.1.1 on 2024-12-10 16:33

import bloomerp.models.fields
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0053_remove_documenttemplate_standard_document_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="documenttemplatefreevariable",
            name="help_text",
            field=models.CharField(
                blank=True,
                help_text="Help text for the variable that will be shown upon creation.",
                max_length=100,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="documenttemplatefreevariable",
            name="name",
            field=models.CharField(
                help_text="The name of the variable.", max_length=100
            ),
        ),
        migrations.AlterField(
            model_name="documenttemplatefreevariable",
            name="required",
            field=models.BooleanField(
                default=False,
                help_text="Signifies whether the variable is required or not.",
            ),
        ),
        migrations.AlterField(
            model_name="documenttemplatefreevariable",
            name="variable_type",
            field=models.CharField(
                choices=[
                    ("date", "Date"),
                    ("boolean", "Boolean"),
                    ("text", "Text"),
                    ("list", "List"),
                    ("integer", "Integer"),
                    ("float", "Decimal"),
                    ("model", "Model"),
                ],
                help_text="The type of the variable.",
                max_length=10,
            ),
        ),
        migrations.AlterField(
            model_name="documenttemplatestyling",
            name="name",
            field=models.CharField(
                help_text="Name of the document template styling.", max_length=100
            ),
        ),
        migrations.AlterField(
            model_name="sqlqueryfilter",
            name="query",
            field=bloomerp.models.fields.CodeField(
                help_text="SQL Query to execute", language="sql"
            ),
        ),
    ]



# myapp/management/commands/list_urls.py

from django.core.management.base import BaseCommand
from django.urls import get_resolver, URLPattern, URLResolver

class Command(BaseCommand):
    help = 'Saves all URLs with their paths and names to a text file'

    def handle(self, *args, **kwargs):
        url_patterns = get_resolver().url_patterns
        with open('urls.txt', 'w') as file:
            self.write_patterns(file, url_patterns)
        self.stdout.write(self.style.SUCCESS('Successfully saved URLs to urls.txt'))

    def write_patterns(self, file, patterns, prefix=''):
        for pattern in patterns:
            if isinstance(pattern, URLPattern):
                name = pattern.name if pattern.name else 'Unnamed'
                file.write(f'{name}: {prefix}{pattern.pattern}\n')
            elif isinstance(pattern, URLResolver):
                self.write_patterns(file, pattern.url_patterns, prefix + str(pattern.pattern))


from django.db.models import Model
from django.contrib.contenttypes.models import ContentType
from bloomerp.models.core import ApplicationField, BloomerpModel
from django.db.models.query import QuerySet
from django.db.models import Q, Model, CharField, TextField
from django.apps import apps


def get_related_models(model:Model, relations:list[str] = ['ForeignKey','ManyToManyField']) -> list[Model]:
    """
    This function returns a list of related models for a given model
    """
    related_models = []
    for field in model._meta.get_fields():
        try:
            if field.get_internal_type() in relations:
                related_models.append(field.related_model)
        except:
            pass
    return related_models 

def get_foreign_occurences_for_model(model:Model) -> list[Model]:
    """
    This function returns a list of models that have a foreign key relationship with the given model
    """
    foreign_occurences = []
    content_types = ContentType.objects.all()
    for content_type in content_types:
        if content_type.model_class():
            related_models = get_related_models(content_type.model_class())
            if model in related_models:
                foreign_occurences.append(content_type.model_class())
    return foreign_occurences

def get_attribute_name_for_foreign_key(model:Model, related_model:Model) -> str:
    """
    This function returns the attribute name of the foreign key field of the related model in the model
    """
    for field in model._meta.get_fields():
        try:
            
            if field.many_to_one and field.related_model == related_model:
                return (field.name, "many_to_one")
            if field.many_to_many and field.related_model == related_model:
                return (field.name, "many_to_many")

        except:
            pass

    return None, None

def get_file_fields_dict_for_model(model:Model) -> list[dict[str, bool]]:
    """
    This function returns a list of file fields for a given model.

    Example:
    file_fields_dict_for_model(User) -> [{'name': 'profile_picture', 'allowed_extensions': ['__all__'], 'required': True}]

    """
    file_fields = []
    for field in model._meta.get_fields():
        try:
            if field.get_internal_type() == 'FileField':
                file_fields.append(
                    {"name": field.name, 
                     "allowed_extensions": ['__all__'],
                     "required": not field.blank 
                    }
                )
            elif field.get_internal_type() == 'BloomerpFileField':
                file_fields.append(
                    {
                        "name": field.name,
                        "allowed_extensions": field.allowed_extensions,
                        "required": not field.blank
                    }
                )
        except:
            pass
    return file_fields

def model_name_singular_slug(model:Model) -> str:
    """
    This function returns the model name in a slug format.

    Example:
    
    model_name_slug(User) -> 'user'
    model_name_slug(UserProfile) -> 'user-profile'

    """
    return model._meta.verbose_name_plural.replace(' ', '-')

def model_name_plural_slug(model:Model) -> str:
    """
    This function returns the model name in a plural slug format.

    Example:

    model_name_plural_slug(User) -> 'users'
    model_name_plural_slug(UserProfile) -> 'user-profiles'
    """
    return model._meta.verbose_name_plural.replace(' ', '-').lower()

def model_name_singular_underline(model:Model) -> str:
    """
    This function returns the model name in a singular underline format.

    Example:

    model_name_singular_underline(User) -> 'user'
    model_name_singular_underline(UserProfile) -> 'user_profile'
    """
    return model._meta.verbose_name.replace(' ', '_')

def model_name_plural_underline(model:Model) -> str:
    """
    This function returns the model name in a plural underline format.

    Example:

    model_name_plural_underline(User) -> 'users'
    model_name_plural_underline(UserProfile) -> 'user_profiles'
    """
    return model._meta.verbose_name_plural.replace(' ', '_').lower()

def string_search(cls, query: str):
    '''
    Class method to search in all string fields of the model.
    Returns a QuerySet filtered by the query in all CharField or TextField attributes.

    Can be given to a model as a class method to search in all string fields of the model.

    Usage:
    ```
    model.string_search = classmethod(string_search)

    '''
    # Get all string fields (CharField and TextField) of the model
    if hasattr(cls, 'string_search_fields') and cls.string_search_fields:
        string_fields = cls.string_search_fields
    else:
        string_fields = [
            field.name for field in cls._meta.fields
            if isinstance(field, CharField) or isinstance(field, TextField)
        ]

    # Build a Q object to filter across all string fields
    query_filter = Q()
    for field in string_fields:
        query_filter |= Q(**{f"{field}__icontains": query})

    # Filter the queryset by the query in any of the string fields
    return cls.objects.filter(query_filter)

def string_search_on_qs(qs: QuerySet, query: str):
    '''
    Function to search in all string fields of a QuerySet.
    Returns a QuerySet filtered by the query in all CharField or TextField attributes.

    Usage:
    ```
    qs = MyModel.objects.all()
    qs = string_search_on_qs(qs, 'search query')
    ```
    '''
    # Get the model of the QuerySet
    model = qs.model

    # Check if the model has a string_search_fields attribute
    if hasattr(model, 'string_search_fields') and model.string_search_fields:
        return model.string_search(query)

    else:
        # Get all string fields (CharField and TextField) of the model
        string_fields = [
            field.name for field in model._meta.fields
            if isinstance(field, CharField) or isinstance(field, TextField)
        ]

    # Build a Q object to filter across all string fields
    query_filter = Q()
    for field in string_fields:
        query_filter |= Q(**{f"{field}__icontains": query})

    # Filter the queryset by the query in any of the string fields
    return qs.filter(query_filter)

def get_bloomerp_file_fields_for_model(model:Model, output='queryset') -> QuerySet[ApplicationField] | list[str]:
    """
    This function returns a QuerySet of ApplicatonFields for a given model containing BloomerpFileFields.
    """
    content_type = ContentType.objects.get_for_model(model)
    qs = ApplicationField.objects.filter(
        content_type=content_type,
        field_type='BloomerpFileField')
    
    if output == 'queryset':
        return qs
    elif output == 'list':
        return list(qs.values_list('field', flat=True))
    
def get_foreign_key_fields_for_model(model:Model) -> QuerySet[ApplicationField]:
    """
    This function returns a QuerySet of ApplicatonFiels for a given model.
    """
    content_type = ContentType.objects.get_for_model(model)
    return ApplicationField.objects.filter(
        content_type=content_type,
        field_type='ForeignKey')

# ---------------------------------
# URL Related Functions
# ---------------------------------
def get_list_view_url(model:Model, type='relative') -> str:
    """
    This function returns the list view url for a given model.
    """
    if type == 'relative':
        return model_name_plural_underline(model) + '_list'
    elif type == 'absolute':
        return model_name_plural_slug(model) + '/list/'

def get_create_view_url(model:Model, type='relative') -> str:
    """
    This function returns the create view url for a given model.

    Example:
        get_create_view_url(User) -> 'users_add'

        get_create_view_url(User, type='absolute') -> 'users/add/'
    """
    if type == 'relative':
        return model_name_plural_underline(model) + '_add'
    elif type == 'absolute':
        return model_name_plural_slug(model) + '/add/'
    
def get_model_dashboard_view_url(model:Model, type='relative') -> str:
    """
    This function returns the dashboard url for a given model.

    Example:
        get_model_dashboard_view_url(User) -> 'users_dashboard'

        get_model_dashboard_view_url(User, type='absolute') -> 'users/dashboard/'
    """
    if type == 'relative':
        return model_name_plural_underline(model) + '_dashboard'
    elif type == 'absolute':
        return model_name_plural_slug(model) + '/'
    
def get_update_view_url(model:Model, type='relative') -> str:
    """
    This function returns the update view url for a given model.

    Example:
        get_update_view_url(User) -> 'users_update'

        get_update_view_url(User, type='absolute') -> 'users/<int_or_uuid:pk>/update/'
    """
    if type == 'relative':
        return model_name_plural_underline(model) + '_detail_update'
    elif type == 'absolute':
        return model_name_plural_slug(model) + '/<int_or_uuid:pk>/update/'
    
def get_detail_view_url(model, type='relative') -> str:
    """
    This function returns the detail view url for a given model.

    Example:
        get_detail_view_url(User) -> 'users_detail'

    """
    if type == 'relative':
        return model_name_plural_underline(model) + '_detail_overview'
    elif type == 'absolute':
        return model_name_plural_slug(model) + '/<int_or_uuid:pk>/'

def get_detail_base_view_url(model, type='relative') -> str:
    """
    This function returns the detail view url for a given model.

    Example:
        get_detail_view_url(User) -> 'users_detail'

    """
    if type == 'relative':
        return model_name_plural_underline(model) + '_detail'

def get_bulk_upload_view_url(model:Model, type='relative') -> str:
    """
    This function returns the bulk upload view url for a given model.

    Example:
        get_bulk_upload_view_url(User) -> 'users_bulk_upload
    """
    if type == 'relative':
        return model_name_plural_underline(model) + '_bulk_upload'
    
def get_base_model_route(model:Model, include_slash=True) -> str:
    """
    This function returns the absolute base route for a given model.

    Example:
        get_base_model_route(User) -> 'users/'
        get_base_model_route(UserProfile) -> 'user-profiles/'

    """
    if include_slash:
        return model_name_plural_slug(model) + '/'
    else:
        return model_name_plural_slug(model)
    
def get_document_template_list_view_url(model: Model, type='relative') -> str:
    """
    This function returns the document template list view url for a given model.

    Example:
        get_document_template_list_view_url(Employee) -> 'employees_document_template_list
        get_document_template_list_view_url(Employee, type='absolute') -> 'employees/document-templates/list/'

    """
    if type == 'relative':
        return model_name_plural_underline(model) + '_document_template_list'
    elif type == 'absolute':
        return model_name_plural_slug(model) + '/document-templates/list/'
    
def get_document_template_generate_view_url(model: Model, type='relative') -> str:
    """
    This function returns the document template generate view url for a given model.

    Example:
        get_document_template_generate_view_url(Employee) -> 'employees_document_template_generate
        get_document_template_generate_view_url(Employee, type='absolute') -> 'employees/<int_or_uuid:pk>/document-templates/<int:template_id>/generate/'

    """
    if type == 'relative':
        return model_name_plural_underline(model) + 'detail_document_templates_generate'
    elif type == 'absolute':
        return model_name_plural_slug(model) + '/<int_or_uuid:pk>/document-templates/<int:template_id>/generate/'
    
def get_model_foreign_key_view_url(model:Model, foreign_model:Model, type='relative') -> str:
    """
    This function returns the foreign key view url for a given model.

    Example:
        get_model_foreign_key_view_url(User, UserProfile) -> 'users_detail_user_profiles'
    """
    if type == 'relative':
        return model_name_plural_underline(model) + '_detail_' + model_name_plural_slug(foreign_model)
    if type == 'absolute':
        return model_name_plural_slug(model) + '/<int_or_uuid:pk>/' + model_name_plural_slug(foreign_model) + '/'
    

# ---------------------------------
# Search models
# ---------------------------------
def search_content_types_by_query(query:str) -> list[ContentType]:
    """
    This function returns a list of ContentTypes that contain the query in their verbose_name_plural.
    """
    content_types = ContentType.objects.all()
    search_results = []

    for content_type in content_types:
        if query.lower() in content_type.model_class()._meta.verbose_name_plural.lower():
            search_results.append(content_type)
    return search_results

def search_objects_by_content_types(query:str, content_types:list[ContentType], limit:int, user=None) -> list[dict[str, QuerySet]]:
    '''Searches objects via content types
    
    Args:
        query: the given query
        content_types: list of content types which should be queries
        limit: limit results to this number of objects

    Returns:
        list of dictionaries containing the model name and the matching objects
    '''
    results = []

    for content_type in content_types:
        model = content_type.model_class()

        if model == ContentType:
            continue

        try:
            if not user.has_perm(f'{model._meta.app_label}.view_{model._meta.model_name}'):
                continue
            
            if issubclass(model, BloomerpModel) and getattr(model, 'allow_string_search', False):
                    # Perform string search using the static method
                matching_objects = model.string_search(query)
            else:
                if hasattr(model, 'allow_string_search') and not model.allow_string_search:
                    continue

                model.string_search = classmethod(string_search)

                matching_objects = model.string_search(query)
            
            # If there are matching objects, add them to the results
            if matching_objects.exists() and limit:
                if len(matching_objects) > limit:
                    matching_objects = matching_objects[0:limit]

                results.append({
                    "model_name": model._meta.verbose_name_plural,
                    "objects": matching_objects
                })
        except:
            pass

    return results
                





# ---------------------------------
# OTHER
# ---------------------------------
def get_initials(object:Model) -> str:
    """
    This function returns the initials of the object.
    """
    return ''.join([word[0].upper() for word in object.__str__().split()])[0:2]


# Generated by Django 5.1.1 on 2024-11-20 17:18

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0047_rename_comment_comment_content_remove_comment_user"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="userlistviewpreference",
            name="is_main",
        ),
    ]



# Generated by Django 5.1.1 on 2024-10-06 13:37

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("test_app", "0014_accommodation_created_by_accommodation_updated_by_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="employee",
            name="date_of_birth",
            field=models.DateField(blank=True, null=True),
        ),
    ]




def parse_bool_parameter(value, default_value=False) -> bool:
    '''
    Function that will parse a string to a boolean value.
    '''
    try:
        if type(value) == bool:
            return value
        
        if value.lower() in ['true', '1']:
            return True
        elif value.lower() in ['false', '0']:
            return False
        else:
            return default_value
    except:
        return default_value    
    


"""
Django settings for test_environment project.

Generated by 'django-admin startproject' using Django 5.1.1.

For more information on this file, see
https://docs.djangoproject.com/en/5.1/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/5.1/ref/settings/
"""
import os
from pathlib import Path
from bloomerp.config import BLOOMERP_APPS, BLOOMERP_MIDDLEWARE, BLOOMERP_USER_MODEL
# Build paths inside the project like this: BASE_DIR / 'subdir'.
BASE_DIR = Path(__file__).resolve().parent.parent


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/5.1/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "django-insecure-z_3c&9yw1%dgom54h6vjkas=1nubb0vvr8zv@!^jp236xbuwh_"

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = []

X_FRAME_OPTIONS = "SAMEORIGIN"


# Application definition
# ---------------------------------
# START GENERATE
# ---------------------------------
LOGIN_URL = "login"
LOGOUT_REDIRECT_URL = '/'  # Redirect to homepage after logout
# ---------------------------------
# END GENERATE
# ---------------------------------

INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "test_app",
    "django_extensions"
]

# ---------------------------------
# START REQUIRED BLOOMERP SETTINGS
# ---------------------------------
INSTALLED_APPS += BLOOMERP_APPS

AUTH_USER_MODEL = 'bloomerp.User'

CRISPY_ALLOWED_TEMPLATE_PACKS = "bootstrap5"

CRISPY_TEMPLATE_PACK = "bootstrap5"
# ---------------------------------
# END REQUIRED BLOOMERP SETTINGS
# ---------------------------------

# Needs to be explicitly set to allow for large bulk uploads
DATA_UPLOAD_MAX_NUMBER_FIELDS = 10240

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

MIDDLEWARE += BLOOMERP_MIDDLEWARE

ROOT_URLCONF = "test_environment.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

WSGI_APPLICATION = "test_environment.wsgi.application"


# Database
# https://docs.djangoproject.com/en/5.1/ref/settings/#databases

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": BASE_DIR / "db.sqlite3",
    }
}


# Password validation
# https://docs.djangoproject.com/en/5.1/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
    },
]


# Internationalization
# https://docs.djangoproject.com/en/5.1/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = True

USE_TZ = True

from django.utils.translation import gettext_lazy as _
LANGUAGES = [
    ("en", _("English")),
    ("es", _("Spanish")),
    ("fr", _("French")),
    ("de", _("German")),
    ("it", _("Italian")),
    ("pt", _("Portuguese")),
    ("nl", _("Dutch")),
]

# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/5.1/howto/static-files/

STATIC_URL = "static/"

# Absolute filesystem path where uploaded media files will be stored
MEDIA_ROOT = os.path.join(BASE_DIR, 'media/')

# URL that handles the media served from MEDIA_ROOT
MEDIA_URL = '/media/'

# Default primary key field type
# https://docs.djangoproject.com/en/5.1/ref/settings/#default-auto-field

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"


# Custom settings for bloomerp

# ---------------------------------
# START GENERATE
BLOOMERP_SETTINGS = {
    "globals" : {
        "organization_name": "My Company",
        "organization_logo": "logo.png",
        "organization_email": "",
    },
    "BASE_URL": "",
    "ROUTERS" : [
        "test_app.views.router"
    ],
    "OPENAI_API_KEY" : "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"
}
# END GENERATE
# ---------------------------------




from bloomerp.utils.router import route
from django.shortcuts import render
from django.http import HttpResponse, HttpRequest
from bloomerp.utils.sql import SqlQueryExecutor # For some reason if I import utils.sql, the router doesn't work
from django.contrib.auth.decorators import login_required

@login_required
@route('execute_sql_query')
def execute_sql_query(request:HttpRequest) -> HttpResponse:
    '''
    Executes a SQL query and returns the result in a html table.
    '''
    # Permission check
    if not request.user.has_perm('bloomerp.execute_sql_query'):
        return HttpResponse('User does not have permission to execute SQL queries')

    query = request.GET.get('sql_query')

    if not query:
        return HttpResponse('No SQL Query provided!')

    executor = SqlQueryExecutor()

    try:
        df = executor.execute_to_df(query)
    except Exception as e:
        return HttpResponse(f'SQL Query error: {e}')
    

    context = {
        "column_names": df.columns.tolist(),
        "data": df.values.tolist()	
    }

    return render(request, 'components/execute_sql_query.html', context)


from bloomerp.utils.router import route
from django.shortcuts import render
from django.http import HttpResponse, HttpRequest
from bloomerp.forms.core import BloomerpModelForm
from bloomerp.models import Todo, User
from django.forms import modelform_factory
from django.utils import timezone
from django.contrib.contenttypes.models import ContentType
from django.contrib.auth.decorators import login_required

@login_required
@route('todos')
def todos(request: HttpRequest) -> HttpResponse:
    """
    Todo component that renders a form to create a new todo item.
    """
    # Get content type and object id from the query string
    content_type_id = request.GET.get('content_type_id')
    object_id = request.GET.get('object_id')

    # Get the filter
    filter_class_dict = {
        'all':'',
        'completed':'',
        'uncompleted':''
    }
    filter = request.GET.get('filter', 'all')

    # Override the filter class
    try:
        filter_class_dict[filter] = 'underline active'
    except KeyError:
        filter_class_dict['all'] = 'underline active'

    # Get the object from the content type and object id
    if content_type_id and object_id:
        content_type = ContentType.objects.get(pk=content_type_id)
        object = content_type.get_object_for_this_type(pk=object_id)
    else:
        object = None


    if request.method == 'POST':
        if request.POST.get('completed_todo_id'):
            todo_id = request.POST.get('completed_todo_id')
            completed_todo = Todo.objects.get(pk=todo_id)
            if completed_todo.is_completed:
                completed_todo.is_completed = False
                completed_todo.datetime_completed = None
            else:
                completed_todo.is_completed = True
                completed_todo.datetime_completed = timezone.now()
            completed_todo.save()

        elif request.POST.get('delete_todo_id'):
            todo_id = request.POST.get('delete_todo_id')
            Todo.objects.get(pk=todo_id).delete()
        else:
            form = initialize_form(request.user, data=request.POST)
            if form.is_valid():
                form.save()
        # Reinitialize form after POST processing
        form = initialize_form(request.user, content_type_id, object_id)
    else:  # GET request
        form = initialize_form(request.user, content_type_id, object_id)

    todos = filter_todos(request.user, content_type_id, object_id)

    # Apply filter on todos
    if filter == 'completed':
        todos = todos.filter(is_completed=True)
    elif filter == 'uncompleted':
        todos = todos.filter(is_completed=False)

    assigned_todos = todos.exclude(assigned_to=request.user)
    my_todos = todos.filter(assigned_to=request.user)

    context = {
        'my_todos': my_todos,
        'assigned_todos': assigned_todos,
        'form': form,
        'content_type_id': content_type_id,
        'object_id': object_id,
        'object': object,
        'filter_class_dict': filter_class_dict
    }
    return render(request, 'components/todos.html', context)



def filter_todos(user: User, content_type_id: int, object_id: int):
    """
    Filter todos based on the user, content type, and object id.
    """
    my_todos = Todo.objects.filter(assigned_to=user) | Todo.objects.filter(requested_by=user)
    if content_type_id and object_id:
        my_todos = my_todos.filter(content_type=content_type_id, object_id=object_id)
    
    return my_todos.order_by('is_completed', '-priority')


def initialize_form(user, content_type_id=None, object_id=None, data=None):
        """Helper function to create and initialize the form."""
        initial = {'assigned_to': user, 'requested_by' : user}
        if content_type_id and object_id:
            initial.update({'content_type': int(content_type_id), 'object_id': int(object_id)})
        Form = modelform_factory(Todo, form=BloomerpModelForm, exclude=['completed', 'datetime_completed'])
        return Form(model=Todo, user=user, initial=initial, data=data)


# Generated by Django 5.1.1 on 2024-11-20 22:22

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0049_alter_file_options_alter_filefolder_options_and_more"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="applicationfield",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "bulk_change",
                    "bulk_delete",
                    "bulk_add",
                ),
                "managed": True,
            },
        ),
        migrations.AlterModelOptions(
            name="bookmark",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "bulk_change",
                    "bulk_delete",
                    "bulk_add",
                ),
                "managed": True,
            },
        ),
        migrations.AlterModelOptions(
            name="comment",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "bulk_change",
                    "bulk_delete",
                    "bulk_add",
                ),
                "managed": True,
            },
        ),
        migrations.AlterModelOptions(
            name="documenttemplate",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "bulk_change",
                    "bulk_delete",
                    "bulk_add",
                ),
                "managed": True,
            },
        ),
        migrations.AlterModelOptions(
            name="documenttemplatefreevariable",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "bulk_change",
                    "bulk_delete",
                    "bulk_add",
                ),
                "managed": True,
            },
        ),
        migrations.AlterModelOptions(
            name="documenttemplateheader",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "bulk_change",
                    "bulk_delete",
                    "bulk_add",
                ),
                "managed": True,
            },
        ),
        migrations.AlterModelOptions(
            name="documenttemplatestyling",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "bulk_change",
                    "bulk_delete",
                    "bulk_add",
                ),
                "managed": True,
            },
        ),
        migrations.AlterModelOptions(
            name="link",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "bulk_change",
                    "bulk_delete",
                    "bulk_add",
                ),
                "managed": True,
            },
        ),
        migrations.AlterModelOptions(
            name="sqlquery",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "bulk_change",
                    "bulk_delete",
                    "bulk_add",
                ),
                "managed": True,
                "verbose_name": "SQL Query",
                "verbose_name_plural": "SQL Queries",
            },
        ),
        migrations.AlterModelOptions(
            name="sqlqueryfilter",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "bulk_change",
                    "bulk_delete",
                    "bulk_add",
                ),
                "managed": True,
                "verbose_name": "SQL Query Filter",
                "verbose_name_plural": "SQL Query Filters",
            },
        ),
        migrations.AlterModelOptions(
            name="user",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "bulk_change",
                    "bulk_delete",
                    "bulk_add",
                )
            },
        ),
        migrations.AlterModelOptions(
            name="userdetailviewpreference",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "bulk_change",
                    "bulk_delete",
                    "bulk_add",
                ),
                "managed": True,
            },
        ),
        migrations.AlterModelOptions(
            name="userdetailviewtab",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "bulk_change",
                    "bulk_delete",
                    "bulk_add",
                ),
                "managed": True,
            },
        ),
        migrations.AlterModelOptions(
            name="userlistviewpreference",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "bulk_change",
                    "bulk_delete",
                    "bulk_add",
                ),
                "managed": True,
            },
        ),
        migrations.AlterModelOptions(
            name="workspace",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "bulk_change",
                    "bulk_delete",
                    "bulk_add",
                ),
                "managed": True,
            },
        ),
    ]



from bloomerp.utils.router import route
from django.shortcuts import render
from django.http import HttpResponse, HttpRequest
from bloomerp.utils.requests import parse_bool_parameter
import uuid
from django.contrib.auth.decorators import login_required
from django.contrib.contenttypes.models import ContentType
from bloomerp.models import ApplicationField

@login_required
@route('datatable_and_filter')
def datatable_and_filter(request:HttpRequest) -> HttpResponse:
    content_type_id = request.GET.get('content_type_id', None)
    include_actions = parse_bool_parameter(request.GET.get('include_actions'), True)
    datatable_id = request.GET.get('datatable_id', 'datatable-' + str(uuid.uuid4()))


    content_type = ContentType.objects.get(pk=content_type_id)
    model = content_type.model_class()

    # Get the application fields
    application_fields = ApplicationField.objects.filter(content_type=content_type)


    if not content_type_id:
        return HttpResponse('content_type_id is required', status=400)
    
    context = {
        'content_type_id': content_type_id,
        'include_actions': include_actions,
        'datatable_id': datatable_id,
        'application_fields': application_fields
    }
    return render(request, 'snippets/datatable_and_filter.html', context)


from django.db import models
from django.contrib.contenttypes.models import ContentType
from bloomerp.models.core import BloomerpModel, ApplicationField
from bloomerp.models.fields import CodeField, TextEditorField, BloomerpFileField, StatusField
from django.utils.translation import gettext_lazy as _
from bloomerp.models import FileFolder

# ---------------------------------
# Document Template Model
# ---------------------------------
class DocumentTemplateHeader(BloomerpModel):
    avatar = None

    class Meta(BloomerpModel.Meta):
        managed = True
        db_table = 'bloomerp_document_template_header'

    name = models.CharField(
        max_length=100,
        blank=False,
        null=False, 
        help_text=_("Name of the template header.")) #Name of the document template header
    header = models.ImageField(
        help_text=_("Image of the header."),
        upload_to='document_templates/headers',
    ) 
    margin_top = models.FloatField(default=0.5, help_text=_("Top margin of the header in inches."))
    margin_bottom = models.FloatField(default=0.0, help_text=_("Bottom margin of the header in inches."))
    margin_left = models.FloatField(default=1.0, help_text=_("Left margin of the header in inches."))
    margin_right = models.FloatField(default=1.0, help_text=_("Right margin of the header in inches."))

    height = models.FloatField(default=1.0, help_text=_("Height of the header in inches."))
    
    def __str__(self):
        return self.name
    
# ---------------------------------
# Document Template Free Variable Model
# ---------------------------------
class DocumentTemplateFreeVariable(BloomerpModel):
    avatar = None

    class Meta(BloomerpModel.Meta):
        managed = True
        db_table = 'bloomerp_document_template_free_variable'

    VARIABLE_TYPE_CHOICES = [
        ('date', 'Date'),
        ('boolean', 'Boolean'),
        ('text', 'Text'),
        ('list', 'List'),
        ('integer', 'Integer'),
        ('float', 'Decimal'),
        ('model','Model')
    ]
    
    name = models.CharField(max_length=100, blank=False, null=False, help_text=_('The name of the variable.')) #Name of the free variable
    help_text = models.CharField(max_length=100, blank=True, null=True, help_text=_('Help text for the variable that will be shown upon creation.')) #Help text for the free variable
    variable_type = models.CharField(
        max_length=10, 
        choices=VARIABLE_TYPE_CHOICES, 
        blank=False, 
        null=False,
        help_text=_('The type of the variable.')
        )
    options = models.TextField(null=True, blank=True)
    required = models.BooleanField(
        null=False, 
        blank=False, 
        default=False,
        help_text=_('Signifies whether the variable is required or not.')
        )

    @property
    def slug(self):
        return self.name.replace(' ','_').lower()
    
    def __str__(self):
        return self.name
    

# ---------------------------------
# Document Template Styling Model
# ---------------------------------
class DocumentTemplateStyling(BloomerpModel):
    avatar = None

    class Meta(BloomerpModel.Meta):
        managed = True
        db_table = 'bloomerp_document_template_styling'

    name = models.CharField(max_length=100, blank=False, null=False, help_text=_("Name of the document template styling."))
    styling = CodeField(language='css', default='') #Content of the styling
    
    def __str__(self):
        return self.name


# ---------------------------------
# Document Template Model
# ---------------------------------
class DocumentTemplate(BloomerpModel):
    class Meta(BloomerpModel.Meta):
        managed = True
        db_table = 'bloomerp_document_template'

    avatar = None

    ORIENTATION_CHOICES = [
        ('portrait', 'Portrait'),
        ('landscape', 'Landscape')
    ]

    PAGE_SIZE_CHOICES = [
        ('A4', 'A4'),
        ('letter', 'Letter'),
        ('A3', 'A3')
    ]

    name = models.CharField(
        max_length=100,
        help_text=_("Name of the document template.")
        ) #Name of the document template
    template = TextEditorField(
        default='Hello world',
        help_text=_("Content of the template, including the variables.")
        ) # Content of the template, including the variables
    model_variable = models.ForeignKey(
        ContentType, 
        on_delete=models.CASCADE,
        help_text=_("Model variable of the document template. Can be used to parse objects from the model into the template."),
        null=True,
        blank=True
        ) # Many to many field to Content Type
    free_variables = models.ManyToManyField(
        DocumentTemplateFreeVariable,
        blank=True,
        null=True,
        help_text=_("A free variable is a variable that is not from a model, and can be inserted in the template at creation time.")
        ) # Many to many field of free variable, a free variable is a variable that is not from a model
    
    template_header = models.ForeignKey(
        DocumentTemplateHeader,
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        help_text=_("Header of the document template.")
        ) #Foreign key to the document template header
    footer = models.TextField(
        help_text=_("Footer content of the document template."),
        blank=True,
        null=True
        )
    styling = models.ForeignKey(
        DocumentTemplateStyling,
        on_delete=models.SET_NULL,
        null=True, 
        blank=True,
        help_text=_("Styling of the document template.")
        ) # Foreign key to the document template styling
    page_orientation = models.CharField(
        max_length=10,
        default='portrait',
        help_text=_("Orientation of the document template."),
        choices=ORIENTATION_CHOICES
        ) # Orientation of the document template
    page_size = models.CharField(
        max_length=10,
        default='A4',
        help_text=_("Size of the document template."),
        choices=PAGE_SIZE_CHOICES
        ) 
    page_margin = models.FloatField(
        default=1.0,
        help_text=_("Margin of the document template in inches.")
        ) # Margin of the document template in inches
    include_page_numbers = models.BooleanField(
        default=True,
        help_text=_("Signifies whether the page numbers are included or not.")
        ) 

    save_to_folder = models.ForeignKey(
        to = FileFolder,
        null=True,
        blank=True,
        help_text=_('Signifies to which folder a file generated from the template needs to be saved upon creation.'),
        on_delete=models.SET_NULL
    )


    form_layout = {
        "General information" : ['name', 'model_variable', 'free_variables'],
        "Template content" : ['template'],
        "Styling" : ['styling', 'template_header','footer', 'page_orientation','page_size','page_margin','include_page_numbers'],
        "Saving" : ['save_to_folder']
    }


    def __str__(self):
        return self.name

    allow_string_search = True
    string_search_fields = ['name']

    def get_related_content_types(model):
        related_content_types = [ContentType.objects.get_for_model(model)]
        return related_content_types

    @property
    def application_fields(self):
        '''
        Returns a queryset of ApplicationField that are related to the model variable of the document template.
        '''
        if self.model_variable is None:
            return ApplicationField.objects.none()
        else:
            qs = ApplicationField.objects.filter(content_type=self.model_variable)
            return qs
    
    def get_variables(self) -> list[(str, str, str)]:
        '''
        Returns a list of tuples with the name and type of the variables in the template.

        The tuple is in the format (name, type, description)
        '''
        variables = []

        # Add the application fields
        for field in self.application_fields:
            variables.append(('object.'+field.field, field.field_type, 'Object field'))

        # Add the free variables
        for variable in self.free_variables.all():
            variables.append((variable.slug, variable.variable_type, 'Free variable'))

        return variables

    @staticmethod
    def get_standard_documents_for_instance(instance):
        content_type = ContentType.objects.get_for_model(instance)
        return DocumentTemplate.objects.filter(model_variable=content_type, standard_document=True)
    


# ---------------------------------
# Signature Request Model
# ---------------------------------

class SignatureRequest:
    '''
    The Signature Request model can be used to request a signature from a (authenticated) user.

    Attributes:
        recipient: The recipient of the signature request.
        recipient_email: The email of the recipient (if the recipient is not provided).
        sender: The sender of the signature request.
        status: The status of the signature request.
        signed_document: The signed document of the signature request.
        signed_at: The date and time when the document was signed.

    '''
    class Meta(BloomerpModel.Meta):
        managed = True
        db_table = 'bloomerp_signature_request'


    avatar = None

    recipient = models.ForeignKey(
        'bloomerp.User',
        on_delete=models.CASCADE,
        help_text=_("Recipient of the signature request."),
        related_name='incoming_signature_requests',
        null=True,
        blank=True
        ) # Foreign key to the recipient
    
    recipient_email = models.EmailField(
        help_text=_("Email of the recipient (if recipient is not provided)."),
        null=True,
        blank=True
        ) # Email of the recipient


    sender = models.ForeignKey(
        'bloomerp.User',
        on_delete=models.CASCADE,
        help_text=_("Sender of the signature request.")
        ) # Foreign key to the sender
    
    status = StatusField(
        default='pending',
        help_text=_("Status of the signature request."),
        colored_choices=[
            ('pending', 'Pending', '#ffcc00'),
            ('processing', 'Processing', '#007bff'),
            ('completed', 'Completed', '#28a745'),
            ('cancelled', 'Cancelled', '#dc3545'),
        ]
    )
    


    signed_document = BloomerpFileField()
    
    signed_at = models.DateTimeField(
        help_text=_("Date and time when the document was signed."),
        blank=True,
        null=True
        ) # Date and time when the document was signed
    
    def __str__(self):
        return f'{self.document_template.name} - {self.recipient.get_full_name()}'

    
    
    



XXXXXXXXXXX
XXXXXXXXXXX
XXXXXXXXX
XXXXXXXXXXXXXXXX
XXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX

XXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXX


# Generated by Django 5.1.1 on 2024-11-05 18:00

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0041_filefolder_content_types"),
    ]

    operations = [
        migrations.RenameModel(
            old_name="Kpi",
            new_name="Widget",
        ),
        migrations.RenameField(
            model_name="userdashboardview",
            old_name="kpis",
            new_name="widgets",
        ),
        migrations.RenameField(
            model_name="userkpidashboardviewpreference",
            old_name="kpi",
            new_name="widget",
        ),
        migrations.AlterUniqueTogether(
            name="userkpidashboardviewpreference",
            unique_together=set(),
        ),
        migrations.AlterField(
            model_name="userkpidashboardviewpreference",
            name="user",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="widget_dashboard_view_preference",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AlterUniqueTogether(
            name="userkpidashboardviewpreference",
            unique_together={("widget", "user")},
        ),
        migrations.AlterModelTable(
            name="userkpidashboardviewpreference",
            table="bloomerp_user_widget_dashboard_view_preference",
        ),
        migrations.AlterModelTable(
            name="widget",
            table="bloomerp_widget",
        ),
    ]



from django import forms

class RichTextEditorWidget(forms.Textarea):
    template_name = 'widgets/tiny_mce_text_editor_widget.html'

    def get_context(self, name, value, attrs):
        attrs = attrs or {}
        attrs.setdefault('id', 'id_%s' % name)
        context = super().get_context(name, value, attrs)
        context['widget'].update({
            'editor_id': attrs['id'],
        })
        return context



# Generated by Django 5.1.1 on 2024-12-29 18:48

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("test_app", "0027_alter_order_status"),
    ]

    operations = [
        migrations.AddField(
            model_name="employee",
            name="gender",
            field=models.CharField(
                blank=True,
                choices=[("M", "Male"), ("F", "Female"), ("O", "Other")],
                max_length=1,
                null=True,
            ),
        ),
    ]



from rest_framework import viewsets
from django_filters import rest_framework as filters
from rest_framework.permissions import IsAuthenticated, BasePermission

class ViewPermission(BasePermission):
    def has_permission(self, request, view):
        return request.user.has_perm(f'{view.model._meta.app_label}.view_{view.model._meta.model_name}')

class AddPermission(BasePermission):
    def has_permission(self, request, view):
        return request.user.has_perm(f'{view.model._meta.app_label}.add_{view.model._meta.model_name}')

class ChangePermission(BasePermission):
    def has_permission(self, request, view):
        return request.user.has_perm(f'{view.model._meta.app_label}.change_{view.model._meta.model_name}')

class DeletePermission(BasePermission):
    def has_permission(self, request, view):
        return request.user.has_perm(f'{view.model._meta.app_label}.delete_{view.model._meta.model_name}')

class BloomerpModelViewSet(viewsets.ModelViewSet):
    # The model will be injected dynamically when the viewset is initialized
    queryset = None
    serializer_class = None
    filter_backends = (filters.DjangoFilterBackend,)
    filterset_fields = '__all__'
    permission_classes = (IsAuthenticated,)

    def get_permissions(self):
        if self.action == 'list' or self.action == 'retrieve':
            self.permission_classes = [IsAuthenticated, ViewPermission]
        elif self.action == 'create':
            self.permission_classes = [IsAuthenticated, AddPermission]
        elif self.action in ['update', 'partial_update']:
            self.permission_classes = [IsAuthenticated, ChangePermission]
        elif self.action == 'destroy':
            self.permission_classes = [IsAuthenticated, DeletePermission]
        return super().get_permissions()

    def get_queryset(self):
        return self.model.objects.all()

    def get_serializer_class(self):
        return self.serializer_class



# Generated by Django 5.1.1 on 2024-12-13 18:46

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0054_documenttemplatefreevariable_help_text_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="documenttemplate",
            name="page_margin",
            field=models.FloatField(
                default=1.0, help_text="Margin of the document template in inches."
            ),
        ),
        migrations.AddField(
            model_name="documenttemplate",
            name="page_orientation",
            field=models.CharField(
                choices=[("portrait", "Portrait"), ("landscape", "Landscape")],
                default="portrait",
                help_text="Orientation of the document template.",
                max_length=10,
            ),
        ),
        migrations.AddField(
            model_name="documenttemplate",
            name="page_size",
            field=models.CharField(
                choices=[("A4", "A4"), ("letter", "Letter"), ("A3", "A3")],
                default="A4",
                help_text="Size of the document template.",
                max_length=10,
            ),
        ),
        migrations.AddField(
            model_name="documenttemplateheader",
            name="height",
            field=models.FloatField(
                default=1.0, help_text="Height of the header in inches."
            ),
        ),
        migrations.AddField(
            model_name="documenttemplateheader",
            name="margin_bottom",
            field=models.FloatField(
                default=0.0, help_text="Bottom margin of the header in inches."
            ),
        ),
        migrations.AddField(
            model_name="documenttemplateheader",
            name="margin_left",
            field=models.FloatField(
                default=1.0, help_text="Left margin of the header in inches."
            ),
        ),
        migrations.AddField(
            model_name="documenttemplateheader",
            name="margin_right",
            field=models.FloatField(
                default=1.0, help_text="Right margin of the header in inches."
            ),
        ),
        migrations.AddField(
            model_name="documenttemplateheader",
            name="margin_top",
            field=models.FloatField(
                default=0.5, help_text="Top margin of the header in inches."
            ),
        ),
        migrations.AlterField(
            model_name="documenttemplate",
            name="footer",
            field=models.TextField(
                blank=True,
                help_text="Footer content of the document template.",
                null=True,
            ),
        ),
    ]



from django.core.management.base import BaseCommand
from django.apps import apps
from django.contrib.contenttypes.models import ContentType
from bloomerp.models import ApplicationField
from django.db import models
from django import db

class Command(BaseCommand):
    help = 'Validates the form layouts for all models.'

    def handle(self, *args, **options):
        model_list = apps.get_models()

        incorrect_number = 0

        for Model in model_list:
            if hasattr(Model, 'form_layout'):
                correct, missing_fields = Model._validate_form_layout()

                if not correct:
                    self.stdout.write(self.style.ERROR(f'Form layout for {Model.__name__} is incorrect. Missing fields: {missing_fields}'))
                    incorrect_number += 1

        if incorrect_number == 0:
            self.stdout.write(self.style.SUCCESS('All form layouts are correct!'))
        else:
            self.stdout.write(self.style.ERROR(f'{incorrect_number} form layouts are incorrect!'))
        


from bloomerp.utils.router import route
from django.urls import reverse
from django.shortcuts import render
from django.http import HttpResponse, HttpRequest
from bloomerp.utils.filters import dynamic_filterset_factory
from bloomerp.utils.models import string_search_on_qs
from django.contrib.contenttypes.models import ContentType
from bloomerp.models import User, UserListViewPreference
from bloomerp.utils.models import model_name_plural_underline
from bloomerp.utils.model_io import BloomerpModelIO
from bloomerp.models import ApplicationField
from django.contrib.auth.decorators import login_required
from django.core.signing import dumps, loads

@login_required
@route('datatable')
def datatable(request:HttpRequest) -> HttpResponse:
    """
    Component for rendering a data table.
    The data table can be filtered, searched, and paginated.
    
    Inputs:
        - data_table_content_type_id: The content type id of the model to be rendered in the data table
        - data_table_include_actions: Whether to include actions in the data table (update, delete, etc.)
        - data_table_string_search: A string to search for in the data table
        - data_table_download: If selected, the response will be a CSV or Excel download of the data table
        - data_table_foreign_key_field: The foreign key field to filter the data table by
        - data_table_foreign_key_value: The foreign key value to filter the data table by
    """
    # Get parameters from the request
    user:User = request.user
    content_type_id = request.GET.get('data_table_content_type_id', None)
    include_actions = request.GET.get('data_table_include_actions', True) # Actions will be rendered to the right of the table for each row
    data_table_string_search = request.GET.get('data_table_string_search', None)
    data_table_download = request.GET.get('data_table_download', None)
    data_table_foreign_key_field = request.GET.get('data_table_foreign_key_field', None)
    data_table_foreign_key_value = request.GET.get('data_table_foreign_key_value', None)
    data_table_limit = request.GET.get('data_table_limit',25)
    data_table_order_by = request.GET.get('data_table_order_by', None)
    data_table_bypass_view_permission_value = request.GET.get('data_table_bypass_view_permission', None)
    
    # View type
    data_table_view_type = request.GET.get('data_table_view_type', 'list')
    data_table_calendar_view_fields = request.GET.getlist('data_table_calendar_view_fields', None)

    # Set actions to True if the string is 'true'
    if include_actions in ['true', 'True']:
        include_actions = True
    else:
        include_actions = False
    
    # Get the model based on the content type id
    content_type = ContentType.objects.get(id=content_type_id)
    model = content_type.model_class()


    # Data table limit
    if data_table_limit == 'all':
        data_table_limit = None
    

    # ---------------------
    # PERMISSION CHECK
    # ---------------------
    # Checks if the user has the permission to view the data table
    # Also checks if the user has the bypass view permission
    if not user.has_perm(f'{model._meta.app_label}.view_{model._meta.model_name}'):
        # If the bypass view permission is set do some extra checks
        if data_table_bypass_view_permission_value:
            value_dict : dict = loads(data_table_bypass_view_permission_value)

            # Check if the user has the permission to view the data table
            if value_dict.get('content_type_id') != int(content_type_id):
                return HttpResponse('User does not have permission to view this data table')
            if value_dict.get('user_id') != user.id:
                return HttpResponse('User does not have permission to view this data table')
            # Get the initial query from the value dict and check if it is a parameter in the get request
            att, val = value_dict.get('initial_query').split('=')
            if request.GET.get(att) != val:
                return HttpResponse('User does not have permission to view this data table')
        else:
            return HttpResponse('User does not have permission to view this data table')


    # Init the queryset
    qs = model.objects.all()

    # Filter the queryset by the foreign key field and value
    if data_table_foreign_key_field and data_table_foreign_key_value:
        qs = qs.filter(**{data_table_foreign_key_field: data_table_foreign_key_value})

    # Create the filterset using the dynamic_filterset_factory
    FilterSet = dynamic_filterset_factory(model)
    filterset = FilterSet(request.GET, queryset=qs)
    qs = filterset.qs

    # Get the list view preference for the user
    list_view_preferences = user.get_list_view_preference_for_model(model)
    
    if not list_view_preferences:
        list_view_preferences = UserListViewPreference.generate_default_for_user(user, content_type)

    # Apply string search
    if data_table_string_search:
        # Remove any leading or trailing whitespace
        data_table_string_search = data_table_string_search.strip()

        qs = string_search_on_qs(qs, data_table_string_search)

    # If the download parameter is set, return a CSV or Excel download
    if data_table_download == 'csv' or data_table_download == 'xlsx':
        fields = [pref.application_field.field for pref in list_view_preferences]
        

    if data_table_download == 'csv':
        model_io = BloomerpModelIO(model)
        file_bytes = model_io.export_to_csv(qs, fields)
        response = HttpResponse(file_bytes, content_type='text/csv')
        response['Content-Disposition'] = f'attachment; filename="{model._meta.verbose_name_plural}_data.csv"'
        return response

    elif data_table_download == 'xlsx':
        model_io = BloomerpModelIO(model)
        file_bytes = model_io.export_to_excel(qs, fields)
        response = HttpResponse(file_bytes, content_type='application/vnd.openxmlformats-officedocument.spreadsheetml.sheet')
        response['Content-Disposition'] = f'attachment; filename="{model._meta.verbose_name_plural}_data.xlsx"'
        return response

    # Get the API endpoint for the model
    api_endpoint = reverse(model_name_plural_underline(model)+'-list')

    # Get the application fields
    application_fields = ApplicationField.objects.filter(
        content_type=ContentType.objects.get_for_model(model)
    ).exclude(
        field_type__in=['Property', 'AutoField','OneToManyField']
    )
    
    if data_table_order_by:
        try:
            qs = qs.order_by(data_table_order_by)
        except Exception as e:
            pass
    
    

    # Limit the queryset
    data_table_total_len = qs.count()
    if data_table_limit:
        try:
            data_table_limit = int(data_table_limit)
            qs = qs[:data_table_limit]
        except:
            pass
    data_table_shown_results = qs.count()

    # Get the target from the HTMX request
    data_table_target = request.htmx.target

    # Get the previous request parameters
    request_params = request.GET.urlencode()


    context = {
        'object_list': qs,
        'content_type_id': content_type_id,
        'list_view_preferences': list_view_preferences,
        'include_actions': include_actions,
        'api_endpoint': api_endpoint,
        'application_fields': application_fields,
        'data_table_foreign_key_field': data_table_foreign_key_field,
        'data_table_foreign_key_value': data_table_foreign_key_value,
        'datatable_id' : data_table_target,
        'request_params': request_params,
        'data_table_order_by': data_table_order_by,
        'data_table_shown_results' : data_table_shown_results,
        'data_table_total_results' : data_table_total_len,
        'data_table_view_type' : data_table_view_type,
        'data_table_calendar_view_fields' : data_table_calendar_view_fields,

        # Permissions
        'user_can_add': user.has_perm(f'{model._meta.app_label}.add_{model._meta.model_name}'),
        'user_can_change': user.has_perm(f'{model._meta.app_label}.change_{model._meta.model_name}'),
        'user_can_delete': user.has_perm(f'{model._meta.app_label}.delete_{model._meta.model_name}'),
        'user_can_bulk_change': user.has_perm(f'{model._meta.app_label}.bulk_change_{model._meta.model_name}'),
        'user_can_bulk_delete': user.has_perm(f'{model._meta.app_label}.bulk_delete_{model._meta.model_name}'),
        'user_can_bulk_add': user.has_perm(f'{model._meta.app_label}.bulk_add_{model._meta.model_name}'),
        'user_can_export': user.has_perm(f'{model._meta.app_label}.export_{model._meta.model_name}'),
    }

    return render(request, 'snippets/datatable.html', context)


# Generated by Django 5.1.1 on 2024-09-27 00:28

import bloomerp.models.fields
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0021_userdashboardview_delete_userkpilistviewpreference"),
    ]

    operations = [
        migrations.AlterField(
            model_name="sqlquery",
            name="query",
            field=bloomerp.models.fields.CodeField(
                help_text="SQL Query to execute", language="sql"
            ),
        ),
    ]



from django.core.management.base import BaseCommand
from django.apps import apps
from django.contrib.contenttypes.models import ContentType
from bloomerp.models import ApplicationField
from django.db import models
from django import db

class Command(BaseCommand):
    help = 'Sync properties with @property decorator and fields in a Django model to ApplicationField'

    def handle(self, *args, **options):
        self.stdout.write(self.style.SUCCESS('Starting to sync ApplicationField model'))

        num = 0

        # Get all models in the project
        model_list = apps.get_models()

        for Model in model_list:
            current_field_names = []  # To track existing field names in the model
            Model : models.Model
            try:
                if hasattr(Model, '_meta'):
                    #----------------------------------------------
                    # Processing properties with @property decorator
                    #----------------------------------------------
                    property_list = [
                        {'field_name': attr, 'field_type': 'Property'}
                        for attr in dir(Model)
                        if isinstance(getattr(Model, attr), property)
                    ]
                    

                    #----------------------------------------------
                    # Processing fields in the model
                    #----------------------------------------------
                    field_list = []
                    fields = Model._meta.get_fields()

                    
                    for field in fields:
                        try:
                            # Try-catch block needed to filter out reverse relation fields
                            meta = {}
                            field_type = field.get_internal_type()

                            try:
                                # Get database column for field
                                if hasattr(field, 'db_column') or hasattr(field, 'column'):
                                    if field.db_column is not None:
                                        db_column = field.db_column
                                        db_table = Model._meta.db_table # Save db table
                                        db_field_type = field.db_type(db.connection)
                                    else:
                                        db_column = field.column
                                        db_table = Model._meta.db_table
                                        db_field_type = field.db_type(db.connection)

                                    # Check if the field_type is none
                                    if db_field_type is None:
                                        db_column = None
                                        db_table = None
                                        db_field_type = None
                                    
                            except Exception as e:
                                db_column = None
                                db_table = None
                                db_field_type = None
                                    
                            #----------------------------------------------
                            # Processing many-to-many fields and ForeignKeys
                            #----------------------------------------------
                            if field_type in ['ForeignKey', 'ManyToManyField','BloomerpFileField']:
                                meta['related_model'] = ContentType.objects.get_for_model(field.related_model).pk

                            #----------------------------------------------
                            # Processing one-to-one and many-to-one fields
                            #----------------------------------------------
                            if field.auto_created:
                                meta['auto_created'] = True
                                if field.one_to_one:
                                    field_type = 'OneToOneField'
                                    
                                if field.many_to_one:
                                    field_type = 'ManyToOneField'
                                
                                if field.one_to_many:
                                    field_type = 'OneToManyField'

                            #----------------------------------------------
                            # Processing status field
                            #----------------------------------------------
                            if field_type == 'StatusField':
                                try:
                                    meta['choices'] = field.choices
                                    meta['colored_choices'] = field.colored_choices
                                    meta['colors'] = {choice[0]: choice[2] for choice in field.colored_choices}
                                except:
                                    pass

                            #----------------------------------------------
                            # Processing CharField
                            #----------------------------------------------
                            if field_type == 'CharField':
                                # Check if field has choices
                                if hasattr(field, 'flatchoices'):
                                    meta['choices'] = field.flatchoices


                            field_info = {
                                'field_name': field.name,
                                'field_type': field_type,
                                'meta': meta,
                                'db_column' : db_column,
                                'db_field_type' : db_field_type,
                                'db_table' : db_table
                            }
                            field_list.append(field_info)
                            current_field_names.append(field.name)  # Track field name
                        except:
                            pass

                    content_type_id = ContentType.objects.get_for_model(Model).id
                    all_fields = property_list + field_list

                    # Sync fields to ApplicationField
                    for field_info in all_fields:
                        field_name = field_info['field_name']
                        current_field_names.append(field_name)  # Track property name
                        field_type = field_info['field_type']
                        meta = field_info.get('meta', None)
                        db_field_type = field_info.get('db_field_type')
                        db_column = field_info.get('db_column')
                        db_table = field_info.get('db_table')


                        if meta:
                            if 'related_model' in meta:
                                related_model = ContentType.objects.get(pk=meta['related_model'])
                        else:
                            related_model = None
                        
                        

                        ApplicationField.objects.update_or_create(
                            content_type_id=content_type_id,
                            field=field_name,
                            defaults={
                                'field_type': field_type,
                                'meta': meta,
                                'related_model': related_model,
                                'db_column' : db_column,
                                'db_table' : db_table,
                                'db_field_type' : db_field_type
                            })

                    # Delete stale ApplicationField entries
                    stale_entries = ApplicationField.objects.filter(
                        content_type_id=content_type_id
                    ).exclude(field__in=current_field_names)

                    stale_entries.delete()

            except AttributeError as e:
                self.stderr.write(self.style.ERROR(f"Error processing model {Model.__name__}: {e}"))


        # Remove all ApplicationFields for which the model no longer exists
        content_type_ids = [ct.id for ct in ContentType.objects.all()]
        stale_entries = ApplicationField.objects.exclude(
            content_type_id__in=content_type_ids
        )
        stale_entries.delete()

        self.stdout.write(self.style.SUCCESS('ApplicationField model synced successfully'))
        





# Generated by Django 5.1.1 on 2024-10-05 20:27

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bloomerp", "0030_alter_file_created_by_alter_file_updated_by_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="file",
            name="persisted",
            field=models.BooleanField(default=False),
        ),
    ]



from django.db import models
import ast
from bloomerp.models.core import File
from django.core.exceptions import ValidationError
import os


# ---------------------------------
# Bloomerp File Field
# ---------------------------------
class BloomerpFileField(models.ForeignKey):
    def __init__(self, *args, allowed_extensions=None, **kwargs):
        """
        Initialize BloomerpFileField with a ForeignKey to the File model and optional file type validation.
        `allowed_extensions` specifies allowed file types; if '__all__' or None, all are allowed.
        """
        self.allowed_extensions = allowed_extensions if allowed_extensions is not None else '__all__'
        kwargs['to'] = 'bloomerp.File'
        kwargs['on_delete'] = models.SET_NULL
        kwargs['null'] = True
        kwargs['blank'] = True # Field should always be optional as we dont want any cascading 
        
        super().__init__(*args, **kwargs)

    def formfield(self, **kwargs):
        """
        Specifies the default form field and widget to use with this model field.
        """
        from bloomerp.widgets.bloomerp_file_field_widget import BloomerpFileFieldWidget

        defaults = {
            'widget': BloomerpFileFieldWidget(),
        }

        defaults.update(kwargs)
        return super().formfield(**defaults)
        
    def validate_file_extension(self, file_instance:File):
        """
        Validate the file extension of the file associated with the foreign key.
        """
        # If allowed_extensions is '__all__', no restriction on file types
        if self.allowed_extensions == '__all__':
            return

        # Get the file extension of the associated file
        ext = os.path.splitext(file_instance.file.name)[1].lower()

        # Check if the extension is in the allowed list
        if ext not in self.allowed_extensions:
            allowed_ext_str = ', '.join(self.allowed_extensions)
            raise ValidationError(f'Unsupported file extension. Allowed extensions are: {allowed_ext_str}')

    def clean(self, value, model_instance):
        """
        Perform the validation on the foreign key reference and ensure the file type is allowed.
        """
        # Call the parent clean method to validate the ForeignKey relationship
        value = super().clean(value, model_instance)

        file_instance = File.objects.get(pk=value)

        # Validate the file extension of the linked file object
        if value:  # Ensure that a valid file instance is passed
            self.validate_file_extension(file_instance)

        return value

# ---------------------------------
# Bloomerp Code Field
# ---------------------------------
class CodeField(models.TextField):
    '''
    A custom model field to store code snippets with syntax highlighting.
    '''
    def __init__(self, *args, language='python', **kwargs):
        self.language = language
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        """
        This method tells Django how to serialize the field for migrations.
        """
        name, path, args, kwargs = super().deconstruct()
        kwargs['language'] = self.language
        return name, path, args, kwargs

    def formfield(self, **kwargs):
        """
        Specifies the default form field and widget to use with this model field.
        """
        from django import forms
        from bloomerp.widgets.code_editor_widget import AceEditorWidget  # Import your custom widget

        defaults = {
            'form_class': forms.CharField,
            'widget': AceEditorWidget(language=self.language),
        }
        defaults.update(kwargs)
        return super().formfield(**defaults)

    # Optional: Add custom validation logic if needed
    def validate(self, value, model_instance):
        super().validate(value, model_instance)
        # Example: Add syntax validation for Python code
        if self.language == 'python':
            import ast
            try:
                ast.parse(value)
            except SyntaxError as e:
                raise ValidationError(f"Invalid Python code: {e}")


# ---------------------------------
# Bloomerp Text Editor Field
# ---------------------------------
class TextEditorField(models.TextField):
    '''Use this field to store rich text content with a text editor.'''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    def formfield(self, **kwargs):
        from django import forms
        from bloomerp.widgets.text_editor import RichTextEditorWidget

        defaults = {
            'form_class': forms.CharField,
            'widget': RichTextEditorWidget(),
        }
        defaults.update(kwargs)
        return super().formfield(**defaults)



# ---------------------------------
# Bloomerp Status Field
# ---------------------------------
# ---------------------------------
# Bloomerp Status Field
# ---------------------------------
class StatusField(models.CharField):
    '''
    A status field inherits from CharField and provides a list of choices and colors.
    It is used to represent the status of a particular object and has color highlighting in the UI.

    Required Arguments:
        colored_choices: A list of tuples where each tuple contains a status, a human-readable name, and a color code (hex code).

    Example Usage:
    ```python
    class Task(models.Model):
        status = StatusField(
            max_length=20,
            colored_choices=[
                ('new', 'New', StatusField.BLUE),
                ('in_progress', 'In Progress', StatusField.ORANGE),
                ('completed', 'Completed', StatusField.GREEN),
            ]
        )
    ```
    '''

    RED = '#ff0000'
    GREEN = '#00ff00'
    BLUE = '#0000ff'
    YELLOW = '#ffff00'
    ORANGE = '#ffa500'
    PURPLE = '#800080'
    CYAN = '#00ffff'
    PINK = '#ff69b4'
    GREY = '#808080'
    BLACK = '#000000'
    WHITE = '#ffffff'

    def __init__(
            self,
            colored_choices: list[tuple[str, str, str]], 
            *args, 
            **kwargs):
        # Turn the colored_choices list into a list of choices
        choices = [(choice[0], choice[1]) for choice in colored_choices]

        # Set the color_choices attribute
        self.colored_choices = colored_choices

        # Call the parent class constructor
        kwargs['choices'] = choices
        super().__init__(*args, **kwargs)

    def get_internal_type(self):
        return "StatusField"

    def db_type(self, connection):
        """
        Returns the database type for this field.
        """
        return 'varchar({})'.format(self.max_length)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        kwargs['colored_choices'] = self.colored_choices
        return name, path, args, kwargs
